import numpy as np

poly = np.polynomial.Polynomial


def E(alpha) -> poly:
    """
    Calculate expected value of mixed winer integral E(I_alpha).
    Entries in multindex alpha coresspond to:

    * 0 - time dimension    
    * positive integer - wiener dimension

    Parameters
    ----------
    alpha : array of integers

    Returns
    -------
    numpy.polynomial.Polynomial
        polynomial in time

    """
    if any(a != 0 for a in alpha):
        return poly([0])
    else:
        return poly([1]).integ(len(alpha))


def E2(alpha, beta=None) -> poly:
    """
    Calculate expected value of a product of two mixed winer integrals E(I_alpha*I_beta) or
    E(I_alpha^2) if beta=None.
    Entries in multindices alpha and beta coresspond to:
    
    * 0 - time dimension    
    * positive integer - wiener dimension

    Parameters
    ----------
    alpha : array of integers
        multindex of the first integral
    beta : array of integers or None
        multindex of the second integral

    Returns
    -------
    numpy.polynomial.Polynomial
        polynomial in time

    """
    beta = alpha if beta is None else beta

    if len(alpha) == 0:
        return E(beta)
    if len(beta) == 0:
        return E(alpha)

    # recursion
    w = poly([0])
    if alpha[-1] == beta[-1] != 0:
        w += E2(alpha[:-1], beta[:-1]).integ()
    if alpha[-1] == 0:
        w += E2(alpha[:-1], beta).integ()
    if beta[-1] == 0:
        w += E2(alpha, beta[:-1]).integ()
    return w



import jax.numpy as jnp
import jax


def fill_diagonal(mat, vec):
    (n, _) = mat.shape
    i, j = jnp.diag_indices(n)
    return mat.at[i, j].set(vec)

vectorized_fill_diagonal = jax.vmap(fill_diagonal, in_axes=(0, 0))


# Compare Kloden-Platen (10.3.7), dimension = d, noiseterms = m
# Generate 'steps' stochastic integral increments at once

def get_wiener_integrals(key, steps=1, noise_terms=1, scheme="euler", p=10):
    if noise_terms == 1 or scheme == 'euler':
        dW_scaled = jax.random.normal(key, shape=(steps, noise_terms))
        dI_scaled = 0.5*(dW_scaled**2 - 1)[..., jax.numpy.newaxis]
    
    elif scheme == 'milstein':
        key1, key2, key3, key4, key5 = jax.random.split(key, num=5)
        xi = jax.random.normal(key1, shape=(steps, 1,noise_terms))
        dW_scaled = xi.squeeze()#jax.random.normal(key2, shape=(steps,noise_terms))

        mu = jax.random.normal(key3, shape=(steps, 1,noise_terms))

        eta = jax.random.normal(key4, shape=(steps, p,noise_terms))

        zeta = jax.random.normal(key5, shape=(steps, p,noise_terms))

        rec = 1 / jax.numpy.arange(1, p + 1)  # 1/r vector
        rho = 1 / 12 - (rec ** 2).sum() / (2 * jax.numpy.pi ** 2)

        a = jax.numpy.sqrt(2) * xi + eta

        Imat_nodiag = (
            xi * jax.numpy.transpose(xi, axes=(0, 2, 1)) / 2
            + jax.numpy.sqrt(rho)
            * (mu * jax.numpy.transpose(xi, axes=(0, 2, 1)) - xi * jax.numpy.transpose(mu, axes=(0, 2, 1)))
            + 1.0
            / (2 * jnp.pi)
            * (
                (
                    jnp.expand_dims(a, 2) * jnp.expand_dims(zeta, 3)
                    - jnp.expand_dims(a, 3) * jnp.expand_dims(zeta, 2)
                )
                * rec[:, jax.numpy.newaxis, jax.numpy.newaxis]
            ).sum(axis=1)
        )

        dI_scaled = vectorized_fill_diagonal(
            Imat_nodiag, 0.5 * (xi ** 2 - 1).squeeze()
        )  # Diagonal entries work differently
    
    elif scheme == 'wagner_platen':
        if noise_terms == 1:
            pass
            #tmpU1 = next(self.normal_generator)
            #tmpU2 = next(self.normal_generator)

            #tmpdt = t - t_max
            #tmpdW = tmpU1*math.sqrt(tmpdt)
            #tmpdZ = 0.5*math.pow(tmpdt,3.0/2.0)*(tmpU1 + (1.0 / math.sqrt(3))*tmpU2 )

        else:
            raise NotImplementedError
    
    else:
        raise NotImplementedError

    return {
        'd_w': dW_scaled,
        'd_ww': dI_scaled
    }

if __name__ == '__main__':
    seed = 0
    key = jax.random.PRNGKey(seed)
    get_wiener_integrals(key, steps=3, noise_terms=1)
    get_wiener_integrals(key, steps=3, noise_terms=2)



import numpy as np

class normal:
  def __init__(self, seed=None):
    self.rng = np.random.default_rng(seed)
    self.n = 1
    self.buffer = np.array([])

  #@profile
  def _sample_more(self):
    new_things = self.rng.normal(size=self.n)
    self.buffer = np.concatenate([self.buffer, new_things])
    self.n *= 2
  
  #@profile
  def get_sample(self, size=1, n=None):
    '''
    Return np.array of Gaussians of specified shape

    Parameters
    ----------
    size : int or tuple of integers

    '''
    n = np.prod(size) if n is None else n
    while n > len(self.buffer):
      self._sample_more()
    x = self.buffer[:n]
    self.buffer = self.buffer[n:]
    return x.reshape(size)


  def get_number_of_samples(self, n):
    '''
    Return np.array of Gaussians of specified length

    Parameters
    ----------
    n : int or tuple of integers

    '''
    while n > len(self.buffer):
      self._sample_more()
    x = self.buffer[:n]
    self.buffer = self.buffer[n:]
    # return x.reshape(n) # slow :<
    return x

  #@profile
  def get_single_sample(self):
    '''
    Return a float
    '''
    if len(self.buffer) == 0:
      self._sample_more()
    x = self.buffer[0]
    self.buffer = self.buffer[1:]
    # return x.reshape(n) # slow :<
    return x

  def __iter__(self):
    return self
  
  def __next__(self):
      return self.get_single_sample()



from copy import deepcopy
import jax
import time
import jax.numpy as jnp
import numpy as np
from tqdm import tqdm
from pychastic.sde_problem import SDEProblem
import pychastic.utils
import pychastic.wiener_integral_moments
from functools import wraps
import jax
from pychastic.vectorized_I_generation import get_wiener_integrals
import tqdm
from jax.experimental.host_callback import id_tap

def contract_all(a, b):
    return jax.numpy.tensordot(a, b, axes=len(b.shape))


def tensordot1(a, b):
    return jax.numpy.tensordot(a, b, axes=1)


def tensordot2(a, b):
    return jax.numpy.tensordot(a, b, axes=2)


class SDESolver:
    """
    Produces realisations of stochastic process to ``solve`` method.
    Controls numerical integration features via attributes.

    Parameters
    ----------
    scheme : {'euler', 'commutative_milstein', 'milstein'}, default: 'euler'
         Type of scheme used for integration.
    dt : float
         Step size in fixed-step integration.

    """

    def __init__(
        self,
        scheme="euler",
        dt=0.01,
        dt_adapting_factor=10,
        min_dt=None,
        max_dt=None,
        error_terms=1,
        target_mse_density=1e-2,
        adaptive=False,
    ):
        self.scheme = scheme  # euler | milstein
        self.dt = dt
        self.min_dt = min_dt or self.dt / dt_adapting_factor
        self.max_dt = max_dt or self.dt * dt_adapting_factor
        self.error_terms = error_terms
        self.target_mse_density = target_mse_density
        self.adaptive = adaptive

    def solve_many(self, problem: SDEProblem, n_trajectories=1, seed=0, chunk_size=1, chunks_per_randomization = None, progress_bar = True):
        """
        Solves SDE problem given by ``problem``. Integration parameters are controlled by attribues of ``SDESolver`` object.

        Parameters
        ----------
        problem : SDEProblem
            (Vector) SDE problem to be solved.
        seed : int, optional
            value of seed for PRNG.
        chunk_size: int or None, optional
            Make steps in solver in chunks of `chunk_size` steps.
            If `chunk_size = n` then value at every nth step is returned.
            If `None` then maximal size of chunk is used only final value is returned.
        chunks_per_randomization: int or None, optional
            Sample wiener trajectories once per `chunks_per_randomization` chunks.
            If `chunks_per_randomization = n` then PRNG runs at every nth chunk.
            If `None`, PRNG runs once, at beginning of simulation.
            Smaller values lead to less memory usage. Larger values increase speed.
        progress_bar: True, False
            Display `tqdm` style progress bar during computation.

        Returns
        -------
        dict
            Under following keys in returned `dict` you'll find:

            * ``time_values`` -- (`n_trajectories`, `steps`) jnp.array containing timestamps coresponding to each trajectory.
            * ``solution_values`` -- (`n_trajectories`, `steps`, `problem_dimension`) `jnp.array` containing values of integrated SDE.
            * ``wiener_values`` -- (`n_trajectories`, `steps`, `noise_dimension`) `jnp.array` containing values of Wiener processes driving the SDE.

        Example
        -------
        >>> import pychastic
        >>> import jax.numpy as jnp
        >>> solver = pychastic.sde_solver.SDESolver()
        >>> problem = pychastic.sde_problem.SDEProblem(
        ... lambda x: jnp.array([1/(2*x[0]),0]),       # [1/2r,0]
        ... lambda x: jnp.array([
        ...    [jnp.cos(x[1]),jnp.sin(x[1])],           # cos \phi,      sin \phi
        ...    [-jnp.sin(x[1])/x[0],jnp.cos(x[1])/x[0]] # -sin \phi / r, cos \phi / r
        ... ]),
        ... x0 = jnp.array([1.0,0.0]), # r=1.0, \phi=0.0
        ... tmax=0.02
        ... )
        >>> solution = solver.solve_many(problem,n_trajectories = 100)
        >>> (r,phi) = (solution["solution_values"][:,-1,:]).transpose()
        >>> compare = {"integrated":(r*jnp.array([jnp.cos(phi),jnp.sin(phi)])).T,"exact":solution["wiener_values"][:,-1,:]+problem.x0}
        >>> print(compare["integrated"][0],compare["exact"][0])
            
        """

        assert problem.x0.shape == problem.a(problem.x0).shape
        assert problem.x0.shape[0] == problem.b(problem.x0).shape[0]

        dimension, noise_terms = problem.b(problem.x0).shape

        def L_t(f):
            @wraps(f)
            def wrapped(x):
                b_val = problem.b(x)
                val = tensordot1(jax.jacobian(f)(x), problem.a(x)) + 0.5 * tensordot2(
                    jax.hessian(f)(x), tensordot1(b_val, b_val.T)
                )
                val = jax.numpy.expand_dims(val, -1)
                return val

            return wrapped

        def L_w(f):
            @wraps(f)
            def wrapped(x):
                return tensordot1(jax.jacobian(f)(x), problem.b(x))

            return wrapped

        def L(f, idx):
            for x in idx:
                if x == "t":
                    f = L_t(f)
                elif x == "w":
                    f = L_w(f)
                else:
                    raise ValueError
            return f

        def step(
            x,
            d_t,
            d_w,
            d_ww=jax.numpy.zeros((noise_terms, noise_terms)),
            d_tw=jax.numpy.zeros((1, noise_terms)),
            d_wt=jax.numpy.zeros((noise_terms, 1)),
            d_www=jax.numpy.zeros((noise_terms, noise_terms, noise_terms)),
            scheme="euler",
        ):

            id_ = lambda x: x
            f_t = L(id_, "t")
            f_w = L(id_, "w")
            f_ww = L(id_, "ww")
            f_tw = L(id_, "tw")
            f_wt = L(id_, "wt")
            f_www = L(id_, "www")

            new_x = x
            new_x += (f_t(x)*d_t).squeeze() + contract_all(f_w(x), d_w)
            if scheme == "euler":
                return new_x

            new_x += contract_all(f_ww(x), d_ww)
            if scheme == "milstein":
                return new_x

            new_x += (
                contract_all(f_tw(x), d_tw)
                + contract_all(f_wt(x), d_wt)
                + contract_all(f_www(x), d_www)
            )
            if scheme == "wagner_platen":
                return new_x

        steps_needed = int(problem.tmax / self.dt)
        tmp = chunks_per_randomization or 1
        chunk_size = chunk_size or steps_needed
        number_of_chunks = ((steps_needed // (chunk_size*tmp)) + (1 if steps_needed % (chunk_size*tmp) else 0))*tmp
        chunks_per_randomization = chunks_per_randomization or number_of_chunks

        key = jax.random.PRNGKey(seed)
        
        def scan_func(carry, input_):
            t, x, w = carry
            
            wiener_integrals = input_
            wiener_integrals['d_w'] *= jax.numpy.sqrt(self.dt)
            wiener_integrals['d_ww'] *= self.dt
            
            t += self.dt
            x = step(x, d_t=self.dt, scheme=self.scheme, **wiener_integrals)
            w += wiener_integrals['d_w']
            return (t, x, w), (t, x, w)

        t0 = 0.0
        w0 = jax.numpy.zeros(noise_terms)

        if progress_bar:
            p_bar = tqdm.tqdm(total = number_of_chunks)
            def tap_func(*args,**kwargs):
                p_bar.update()
        else:
            def tap_func(*args,**kwargs):
                pass

        def chunk_function(chunk_start, wieners_chunk):
            # Parameters: chunk_start = (t0, x0, w0) values at beggining of chunk
            #             wieners_chunk = array of wiener increments
            id_tap(tap_func,0)
            z = jax.lax.scan( scan_func , chunk_start , wieners_chunk )[0] #discard trajectory at chunk resolution
            return z, z

        def get_solution_fragment(starting_state,key):
            wiener_integrals = get_wiener_integrals(key, steps=chunk_size*chunks_per_randomization, noise_terms=noise_terms, scheme=self.scheme)    

            last_state , (time_values, solution_values, wiener_values) = jax.lax.scan( 
                chunk_function, 
                starting_state,
                jax.tree_map(lambda x: jnp.reshape(x,(-1,chunk_size)+x.shape[1:]), wiener_integrals)
            ) #discard carry, remember trajectory
            
            return (
                    last_state,
                    dict(
                    time_values=time_values,
                    solution_values=solution_values,
                    wiener_values=wiener_values,
                    )
                   )

        @jax.vmap
        def get_solution(key):
            _ , chunked_solution = jax.lax.scan(
                lambda state, key: get_solution_fragment(state,key),
                (t0,problem.x0,w0),
                jax.random.split(key, number_of_chunks // chunks_per_randomization)
                )

            return jax.tree_map(lambda x: x.reshape((-1,)+x.shape[2:]),chunked_solution) #combine big chunks into one trajectory

        keys = jax.random.split(key, n_trajectories)
        solutions = get_solution(keys)
        if progress_bar: 
            p_bar.refresh()
        return solutions

    def solve(self, problem, seed=0, chunk_size=1, chunks_per_randomization = None, progress_bar = True):
        """
        Solves SDE problem given by ``problem``. Integration parameters are controlled by attribues of ``SDESolver`` object.

        Parameters
        ----------
        problem : SDEProblem
            (Vector) SDE problem to be solved.
        seed : int, optional
            value of seed for PRNG.
        chunk_size: int or None, optional
            Make steps in solver in chunks of `chunk_size` steps.
            If `chunk_size = n` then value at every nth step is returned.
            If `None` then maximal size of chunk is used only final value is returned.
        chunks_per_randomization: int or None, optional
            Sample wiener trajectories once per `chunks_per_randomization` chunks.
            If `chunks_per_randomization = n` then PRNG runs at every nth chunk.
            If `None`, PRNG runs once, at beginning of simulation.
            Smaller values lead to less memory usage. Larger values increase speed.
        progress_bar: True, False
            Display `tqdm` style progress bar during computation.

        Returns
        -------
        dict
            Under following keys in returned `dict` you'll find:

            * ``time_values`` -- (`steps`,) jnp.array containing timestamps coresponding to each trajectory.
            * ``solution_values`` -- (`steps`, `problem_dimension`) `jnp.array` containing values of integrated SDE.
            * ``wiener_values`` -- (`steps`, `noise_dimension`) `jnp.array` containing values of Wiener processes driving the SDE.

        Example
        -------
        >>> import pychastic
        >>> import jax.numpy as jnp
        >>> solver = pychastic.sde_solver.SDESolver()
        >>> problem = pychastic.sde_problem.SDEProblem(
        ... lambda x: jnp.array([1/(2*x[0]),0]),       # [1/2r,0]
        ... lambda x: jnp.array([
        ...    [jnp.cos(x[1]),jnp.sin(x[1])],           # cos \phi,      sin \phi
        ...    [-jnp.sin(x[1])/x[0],jnp.cos(x[1])/x[0]] # -sin \phi / r, cos \phi / r
        ... ]),
        ... x0 = jnp.array([1.0,0.0]), # r=1.0, \phi=0.0
        ... tmax=0.02
        ... )
        >>> solution = solver.solve(problem)
        >>> (r,phi) = (solution["solution_values"][-1,:])
        >>> compare = {"integrated":(r*jnp.array([jnp.cos(phi),jnp.sin(phi)])).T,"exact":solution["wiener_values"][-1,:]+problem.x0}
        >>> print(compare["integrated"],compare["exact"])
            
        """
        solution = self.solve_many(problem, n_trajectories=1, seed=seed)
        solution = jax.tree_map(lambda x: x[0], solution)
        return solution

if __name__ == '__main__':
    a = 1
    b = 1
    scalar_geometric_bm = SDEProblem(
        a = lambda x: a*x,
        b = lambda x: b*x,
        x0 = 1.0,
        tmax = 1.0,
        exact_solution = lambda x0, t, w: x0*np.exp((a-0.5*b*b)*t+b*w)   
    )
    problem = scalar_geometric_bm
    solver = SDESolver()
    steps = 100
    dt = problem.tmax / steps
    solver.dt = dt
    solver.solve_many(problem, n_trajectories=1000)



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(name='pychastic',
      version='0.2.0',
      description='Solvers for stochastic differential equations',
      url='https://github.com/RadostW/stochastic',
      author='Radost Waszkiewicz & Maciej Bartczak',
      author_email='radost.waszkiewicz@gmail.com',
      long_description=long_description,
      long_description_content_type='text/markdown',  # This is important!
      project_urls = {
          'Documentation': 'https://pychastic.readthedocs.io',
          'Source': 'https://github.com/RadostW/stochastic'
      },
      license='MIT',
      install_requires = ['jax','jaxlib','numpy','sortedcontainers','tqdm'],
      packages=['pychastic'],
      zip_safe=False)



import jax
import jax.numpy as jnp


def _is_scalar(x):
  return jax.numpy.isscalar(x) or x.ndim == 0

class SDEProblem:
  '''
  Stores a vector stochastic differential equation of the form:
  :math:`d\\mathbf{x} = \\mathbf{a(x)} dt + \\mathbf{B(x)} d\\mathbf{w}`, where
  ``a`` and ``B`` are vector of drifts and matrix of noise coefficients.

  ``a`` vector and ``B`` matrix should be formatted such that

  .. prompt:: python >>> auto

    >>> x = x + a(x)*dt + np.dot(B(x),dw)

  is a valid code fragment (note ``B`` has as many columns as there are noise terms).

  Parameters
  ----------
  a : callable
    Function describing drift term of the equation, should return np.array of length ``dimension``.
  b : callable
      Function describing noise term of the equation, should return np.array of size ``(dimension,noiseterms)``.
  dimension : int
      Dimension of the space in which :math:`d\\mathbf{x}` sits.
  noiseterms : int
      Dimension of the space in which :math:`d\\mathbf{w}` sits.
  x0 : np.array
      Initial value of the stochastic process.
  tmax : float
      Time at which integration should stop.

  Example
  -------
  >>> import numpy as np
  >>> problem = pychastic.sde_problem.VectorSDEProblem(lambda x: np.array([1,1]), lambda x: np.array([[1,0.5],[0.5,1]]), 2, 2, np.array([1.5,0.5]), 1)

  '''
  def __init__(self, a, b, x0, tmax, exact_solution=None):
    
    if not tmax > 0:
      raise ValueError(f'tmax has to bo posiitve, not {tmax}')
    self.tmax = tmax
    
    # dimension & shape validation
        
    x0 = jnp.array(x0, dtype=jax.numpy.float32)

    (a_x0_shape, a_x0_dtype) = (jax.eval_shape(a,x0).shape,jax.eval_shape(a,x0).dtype)
    (b_x0_shape, b_x0_dtype) = (jax.eval_shape(b,x0).shape,jax.eval_shape(b,x0).dtype)

    if x0.ndim == len(a_x0_shape) == len(b_x0_shape) == 0:
      self.x0 = x0.reshape(1)
      # scalar case
      if len(a_x0_shape) != 1:
        new_a = lambda x: a(x).reshape(1)
        # TODO warn about reshaping
      if len(b_x0_shape) != 2:
        new_b = lambda x: b(x).reshape(1, 1)
        # TODO warn about reshaping
      
      self.dimension = self.noise_terms = 1
      self.a = new_a
      self.b = new_b

    elif x0.ndim == len(a_x0_shape) == 1 and len(b_x0_shape) == 2:
      # vector case
      if not x0.shape[0] == a_x0_shape[0] == b_x0_shape[0]:
        raise ValueError(f'Incosistent shapes: {x0.shape}, {a_x0_shape}, {b_x0_shape}')   

      self.x0 = x0
      self.a = a
      self.b = b
      self.dimension, self.noise_terms = b_x0_shape

    else:
      raise ValueError(f'Inconsistent dimensions: {x0.shape}, {a_x0_shape}, {b_x0_shape}')
      
    

    if not isinstance(x0, jnp.ndarray):
        raise ValueError(f"{x0} should be jnp.array, not {type(x0)}")

    # dtype validation
    for val, key in [(x0.dtype, 'initial conditon'), (a_x0_dtype, 'drift term'), (b_x0_dtype, 'noise term')]:
      if not jnp.issubdtype(x0.dtype, jnp.floating):
        raise ValueError(f"{key} dtype should be float, not {val.dtype}.")

    # exact solution validation
    # TODO
    self.exact_solution = exact_solution

SDEProblem(lambda x: 1, lambda x: 2, 3, 4)



from numpy.core.fromnumeric import transpose
import sortedcontainers
from pychastic.cached_gaussian import normal
import numpy as np
import jax.numpy as jnp
import math

class Wiener:
    '''
    Class for sampling, and memorization of Wiener process.
    '''
    def __init__(self, seed=None):
        self.sample_points = sortedcontainers.SortedDict()
        self.sample_points[0.] = {
            'w': 0.0,
            #'zToPrevPoint': 0.0
        }
        self.normal_generator = normal(seed=seed)
        self.t_max = 0
        self.last_w = 0

    #@profile
    def get_w(self, t):
        '''
        Get value of Wiener probess at specified timestamp.

        Parameters
        ----------
        t: float
            Time at which the process should be sampled. Has to be non-negative.

        Returns
        -------
        float
            Value of Wiener process at time ``t``.

        Example
        -------
        >>> wiener = Wiener()
        >>> dW = wiener.get_w(1.0) - wiener.get_w(0.0)
        >>> dW
        0.321 #random value from N(0,1)


        '''
        if not t >= 0:
            raise ValueError('Illegal (negative?) timestamp')

        if t in self.sample_points:
            return self.sample_points[t]['w']

        t_max = self.t_max
        if t > t_max:
            normal = next(self.normal_generator)
            next_w = self.last_w + np.sqrt(t-t_max)*normal
            self.sample_points[t] = {'w': next_w}
            self.t_max = t
            self.last_w = next_w

        else:
            next_i = self.sample_points.bisect_left(t)
            next_t = self.sample_points.peekitem(next_i)[0]
            prev_t = self.sample_points.peekitem(next_i-1)[0]
            next_w = self.sample_points.peekitem(next_i)[1]['w']
            prev_w = self.sample_points.peekitem(next_i-1)[1]['w']
            w = prev_w + (t-prev_t)/(next_t-prev_t)*(next_w-prev_w) + next(self.normal_generator)*np.sqrt((next_t-t)*(t-prev_t)/(next_t-prev_t))
            assert np.isfinite(w)
            self.sample_points[t] = {'w': w}

        return self.sample_points[t]['w']


    def get_z(self, t1, t2):
        raise NotImplementedError


class WienerWithZ:
    '''
    Class for sampling, and memorization of Wiener process and first nontrivial Stochastic integral.
    '''
    def __init__(self,seed=None):
        self.sample_points = sortedcontainers.SortedDict()
        self.sample_points[0.] = {
            'w': 0.0,
            'zToPrevPoint': 0.0
        }
        self.normal_generator = normal(seed=seed)

    def get_w(self,t):
        '''
        Get value of Wiener probess at specified timestamp.

        Parameters
        ----------
        t: float
            Time at which the process should be sampled. Has to be non-negative.

        Returns
        -------
        float
            Value of Wiener process at time ``t``.

        Example
        -------
        >>> wiener = WienerWithZ()
        >>> dW = wiener.get_w(1.0) - wiener.get_w(0.0)
        >>> dW
        0.321 #random value from N(0,1)


        '''
        self.ensure_sample_point(t)
        return self.sample_points[t]['w']

    def get_z(self,t1,t2):
        '''
        Get value of first nontrivial, primitive stochastic integral I(1,0),
        (Kloden-Platen 10.4.2)

        .. math ::  Z_{t_1}^{t_2} = \int_{t_1}^{t_2} \int_{t_1}^{s_2} dW_{s_1} ds_2

        Parameters
        ----------
        t: float
            Time at which the process should be sampled. Has to be non-negative.

        Returns
        -------
        float
            Value of Wiener process at time ``t``.

        Example
        -------
        >>> wiener = WienerWithZ()
        >>> dZ = wiener.get_z(0.0,0.1)
        >>> dZ
        0.321 #random value from N(0,1)


        '''
        if t1 >= t2:
            raise ValueError

        self.ensure_sample_point(t1)
        self.ensure_sample_point(t2)

        Z = 0
        w1 = self.sample_points[t1]['w']
        it_lower = self.sample_points.irange(t1, t2)
        it_upper = self.sample_points.irange(t1, t2)
        next(it_upper)
        for t_upper in it_upper:
            t_lower = next(it_lower)
            Z += self.sample_points[t_upper]['zToPrevPoint']
            dt = t_upper-t_lower
            dw = self.sample_points[t_lower]['w'] - w1
            Z += dw*dt

        return Z


    def ensure_sample_point(self,t):
        '''
        Ensures ``t`` is in the dictionary of sampled time instances. If not there yet samples new point 
        either to the right of all existing points or inbetween existing points.
        '''
        if t in self.sample_points.keys():
            return

        if t < 0:
            raise ValueError

        t_max = self.sample_points.keys()[-1]
        if t > t_max:
            #Kloden-Platen 10.4.3
            # (4.3)       dW = U1 sqrt(dt), dZ = 0.5 dt^(3/2) (U1 + 1/sqrt(3) U2)
            tmpU1 = next(self.normal_generator)
            tmpU2 = next(self.normal_generator)

            tmpdt = t - t_max
            tmpdW = tmpU1*math.sqrt(tmpdt)
            tmpdZ = 0.5*math.pow(tmpdt,3.0/2.0)*(tmpU1 + (1.0 / math.sqrt(3))*tmpU2 )

            self.sample_points[t] = {'w': self.sample_points[t_max]['w'] + tmpdW, 'zToPrevPoint': tmpdZ}
        else:
            #Somewhere inside sampled points
            next_i = self.sample_points.bisect_left(t)
            next_t = self.sample_points.peekitem(next_i)[0]
            prev_t = self.sample_points.peekitem(next_i-1)[0]
            next_w = self.sample_points.peekitem(next_i)[1]['w']
            prev_w = self.sample_points.peekitem(next_i-1)[1]['w']

            wt1t3 = next_w - prev_w
            zt1t3 = self.sample_points.peekitem(next_i)[1]['zToPrevPoint']

            (t1,t2,t3) = (prev_t,t,next_t)
            i1 = t2-t1
            i2 = t3-t2
            I = t3-t1

            varwt1t2 =  i1*i2*(i1*i1-i1*i2+i2*i2)/(I*I*I);
            cov = i1*i1*i2*i2*(i2-i1)/(2*I*I*I);
            varzt1t2 = i1*i1*i1*i2*i2*i2/(3*I*I*I);

            (wt1t2, zt1t2) = self._DrawCovaried(varwt1t2,cov,varzt1t2)

            #Add conditional mean
            wt1t2 += wt1t3*i1*(i1-2*i2)/(I*I)  + zt1t3*6*i1*i2/(I*I*I)
            zt1t2 += wt1t3*(-1)*i1*i1*i2/(I*I) + zt1t3*i1*i1*(i1+3*i2)/(I*I*I)

            wt2 = prev_w + wt1t2

            #Break Z integration interval into two segments
            self.sample_points[next_t]['zToPrevPoint'] = (
                      self.sample_points.peekitem(next_i)[1]['zToPrevPoint']
                      - zt1t2
                      - (next_t-t)*(wt2-prev_w))
            self.sample_points[t] = {'w' : wt2, 'zToPrevPoint' : zt1t2}

    def _DrawCovaried(self,xx,xy,yy):
        '''
        Draws x,y from N(0,{{xx,xy},{xy,yy}}) distribution
        '''
        (x,y) = self._DrawCorrelated(xy/math.sqrt(xx*yy))
        return (math.sqrt(xx)*x,math.sqrt(yy)*y)

    def _DrawCorrelated(self,cor):
        '''
        Draws correalted normal samples with correlation ``cor``
        '''
        z1 = next(self.normal_generator)
        z2 = next(self.normal_generator)
        return (math.sqrt(1-cor*cor)*z1 + cor*z2,z2)


class VectorWiener:
    '''
    Class for sampling, and memorization of vector valued Wiener process.


    Parameters
    ----------
    noiseterms : int
        Dimensionality of the vector process (i.e. number of independent Wiener processes).

    Example
    -------
    >>> vw = pychastic.wiener.VectorWiener(2)
    >>> vw.get_w(1)
    array([0.21,-0.31]) # random, independent from N(0,1)

    '''
    def __init__(self,noiseterms : int):
        self.sample_points = sortedcontainers.SortedDict()
        self.noiseterms = noiseterms
        self.sample_points[0.] = {
            'w': np.array([0.0 for x in range(0,noiseterms)]),
        }
        self.normal_generator = normal()


    def get_w(self, t):
        '''
        Get value of Wiener probess at specified timestamp.

        Parameters
        ----------
        t: float
            Time at which the process should be sampled. Has to be non-negative.

        Returns
        -------
        np.array
            Value of Wiener processes at time ``t``.

        Example
        -------
        >>> vw = VectorWiener(2)
        >>> dW = vw.get_w(1.0) - vw.get_w(0.0)
        >>> dW
        array([0.321,-0.123]) #random, each from N(0,1)

        '''
        if t < 0:
            raise ValueError('Negative timestamp')

        if t in self.sample_points:
            return self.sample_points[t]['w']

        (t_max, last_values) = self.sample_points.peekitem() # last item is default
        if t > t_max:
            nvec = self.normal_generator.get_number_of_samples(self.noiseterms)
            #nvec = np.array([next(self.normal_generator) for x in range(0,self.noiseterms)]) # slow :<
            w_val = np.array(last_values['w'] + np.sqrt(t-t_max)*nvec)
            self.sample_points[t] = {'w': w_val}
        else:
            #print(f'Called with t {t}, current t_max is {t_max}')
            #print(self.sample_points)
            raise NotImplementedError

        return self.sample_points[t]['w']
    def get_commuting_noise(self, t1, t2):
        '''
        Get value of commutative noise matrix (compare Kloden-Platen (10.3.15))

        Define :math:`I_{jk}` as

        .. math :: I_{jk}(t_1,t_2) = \int_{t_1}^{t_2} \int_{t_1}^{s_1} dW_j(s_2) dW_k(s_1)

        Then for :math:`j \\neq k`

        .. math :: I_{jk} + I_{kj} = \Delta W_j \Delta W_k

        Parameters
        ----------
        t1 : float
            Lower bound of double stochastic integrals
        t2 : float
            Upper bound of double stochastic integrals

        Returns
        -------
        np.array
            Symmetric square matrix `noiseterms` by `noiseterms` containing :math:`I_{jk}` approximants as components.

        '''

        if t1 < 0 or t2 < 0:
            raise ValueError

        if t1 > t2:
            raise ValueError

        if (t1 in self.sample_points) and (t2 in self.sample_points):
            dW = self.sample_points[t2]['w'] - self.sample_points[t1]['w']
            dV = self.sample_points[t2]['w'] - self.sample_points[t1]['w']

            prod = np.outer(dW,dV)
            #halfdiag = np.oneslike(prod) - 0.5*np.identity(self.noiseterms)

            #return prod*halfdiag - (t2-t1)*np.identity(self.noiseterms)
            return 0.5*prod


        t_max = self.sample_points.keys()[-1]
        if t1 > t_max:
            nvec = self.normal_generator.get_sample(self.noiseterms,n=self.noiseterms)
            # nvec = np.array([next(self.normal_generator) for x in range(0,self.noiseterms)]) # slow :<
            self.sample_points[t1] = {'w': self.sample_points[t_max]['w'] + np.sqrt(t1-t_max)*nvec}
        elif t1 not in self.sample_points:
            raise NotImplementedError

        if t2 > t_max:
            nvec = self.normal_generator.get_sample(self.noiseterms,n=self.noiseterms)
            # nvec = np.array([next(self.normal_generator) for x in range(0,self.noiseterms)]) # slow :<
            self.sample_points[t2] = {'w': self.sample_points[t_max]['w'] + np.sqrt(t2-t_max)*nvec}
        elif t2 not in self.sample_points:
            raise NotImplementedError

        dW = self.sample_points[t2]['w'] - self.sample_points[t1]['w']
        dV = self.sample_points[t2]['w'] - self.sample_points[t1]['w']

        prod = np.outer(dW,dV)
        #halfdiag = np.oneslike(prod) - 0.5*np.identity(self.noiseterms)

        #return prod*halfdiag - (t2-t1)*np.identity(self.noiseterms)
        return 0.5*prod

    def get_commuting_noise_component(self, t1, t2, j, k):
        '''
        Get value of commutative noise component (compare Kloden-Platen (10.3.15)).

        Define :math:`I_{jk}` as

        .. math :: I_{jk}(t_1,t_2) = \int_{t_1}^{t_2} \int_{t_1}^{s_1} dW_j(s_2) dW_k(s_1)

        Then for :math:`j \\neq k`

        .. math :: I_{jk} + I_{kj} = \Delta W_j \Delta W_k

        Parameters
        ----------
        t1 : float
            Lower bound of double stochastic integrals
        t2 : float
            Upper bound of double stochastic integrals
        j : int
            Index of the first of Wiener processes
        k : int
            Index of the second of Wiener processes

        Returns
        -------
        float
            Value of stochastic integral with specified time bounds.        

        '''

        if t1 < 0 or t2 < 0:
            raise ValueError

        if t1 > t2:
            raise ValueError

        if (t1 in self.sample_points) and (t2 in self.sample_points):
            dW = self.sample_points[t2]['w'][j] - self.sample_points[t1]['w'][j]
            dV = self.sample_points[t2]['w'][k] - self.sample_points[t1]['w'][k]
            if j != k:
                return 0.5*dW*dV
            else:
                return 0.5*(dW*dW - (t2-t1))


        t_max = self.sample_points.keys()[-1]
        if t1 > t_max:
            #nvec = np.array([next(self.normal_generator) for x in range(0,self.noiseterms)])
            nvec = self.normal_generator.get_sample(self.noiseterms,n=self.noiseterms)
            self.sample_points[t1] = {'w': self.sample_points[t_max]['w'] + np.sqrt(t1-t_max)*nvec}
        elif t1 not in self.sample_points:
            raise NotImplementedError

        if t2 > t_max:
            #nvec = np.array([next(self.normal_generator) for x in range(0,self.noiseterms)])
            nvec = self.normal_generator.get_sample(self.noiseterms,n=self.noiseterms)
            self.sample_points[t2] = {'w': self.sample_points[t_max]['w'] + np.sqrt(t2-t_max)*nvec}
        elif t2 not in self.sample_points:
            raise NotImplementedError

        dW = self.sample_points[t2]['w'][j] - self.sample_points[t1]['w'][j]
        dV = self.sample_points[t2]['w'][k] - self.sample_points[t1]['w'][k]

        if j != k:
            return dW*dV
        else:
            return 0.5*(dW*dW - (t2-t1))


class VectorWienerWithI:
    '''
    Class for sampling, and memorization of vector valued Wiener process including first
    nontrivial vector stochastic integral :math:`I_{jk}`.


    Parameters
    ----------
    noiseterms : int
        Dimensionality of the vector process (i.e. number of independent Wiener processes).
    p : int, default: 10
        Number of terms in fourier expansion of the stochastic integral. 
        The higher the number, the more precise approximation becomes.

    Example
    -------
    >>> vw = pychastic.wiener.VectorWienerWithJ(2)
    >>> vw.get_w(1)
    array([0.21,-0.31]) # random, independent from N(0,1)

    '''
    def __init__(self,noiseterms : int, p = 10):
        self.sample_points = sortedcontainers.SortedDict()
        self.noiseterms = noiseterms
        self.sample_points[0.] = {
            'w': np.zeros(noiseterms),
            'IToPrevPoint' : np.zeros((noiseterms,noiseterms))
        }
        self.normal_generator = normal()
        self.p = p

    def get_w(self, t):
        '''
        Get value of Wiener probess at specified timestamp.

        Parameters
        ----------
        t: float
            Time at which the process should be sampled. Has to be non-negative.

        Returns
        -------
        np.array
            Value of Wiener processes at time ``t``.

        Example
        -------
        >>> vw = VectorWienerWithI(2)
        >>> dW = vw.get_w(1.0) - vw.get_w(0.0)
        >>> dW
        array([0.321,-0.123]) #random, each from N(0,1)

        '''
        if t < 0:
            raise ValueError('Negative timestamp')

        if t in self.sample_points:
            return self.sample_points[t]['w']

        t_max = self.sample_points.keys()[-1]
        if t > t_max:
            self.ensure_sample_point(t)
        else:
            next_i = self.sample_points.bisect_left(t)
            next_t = self.sample_points.peekitem(next_i)[0]
            prev_t = self.sample_points.peekitem(next_i-1)[0]
            if ( next_t - t < 2*jnp.finfo(type(t)).eps ):
                return self.sample_points[next_t]['w']
            elif ( t - prev_t < 2*jnp.finfo(type(t)).eps ):
                return self.sample_points[prev_t]['w']
            else:
                raise NotImplementedError('Conditional subsampling not implemented.')
                # #### TODO #### Implement subsampling

        return self.sample_points[t]['w']

    def get_I_matrix(self, t1, t2):
        '''
        Get value of double integrals :math:`I_{jk}` (compare Kloden-Platen (10.3.5))

        Define :math:`I_{jk}` as

        .. math :: I_{jk}(t_1,t_2) = \int_{t_1}^{t_2} \int_{t_1}^{s_1} dW_j(s_2) dW_k(s_1)

        Parameters
        ----------
        t1 : float
            Lower bound of double stochastic integrals
        t2 : float
            Upper bound of double stochastic integrals

        Returns
        -------
        np.array
            Symmetric square matrix `noiseterms` by `noiseterms` containing :math:`I_{jk}` approximants as components.

        '''

        if not (t1 >= 0 and t2 >= 0):
            raise ValueError('Illegal timestamps for sampling. (Negative?)')

        if t1 > t2:
            raise ValueError('Wrong order of integration limits.')

        t_max = self.sample_points.keys()[-1]
        if t1 > t_max:
            self.ensure_sample_point(t1)
        elif t1 not in self.sample_points:
            next_i = self.sample_points.bisect_left(t1)
            next_t = self.sample_points.peekitem(next_i)[0]
            prev_t = self.sample_points.peekitem(next_i-1)[0]
            if ( next_t - t1 < 2*jnp.finfo(type(t1)).eps ):
                t1 = next_t
            elif ( t1 - prev_t < 2*jnp.finfo(type(t1)).eps ):
                t1 = prev_t
            else:
                raise NotImplementedError('Conditional subsampling not implemented.')

        if t2 > t_max:
            self.ensure_sample_point(t2)
        elif t2 not in self.sample_points:
            next_i = self.sample_points.bisect_left(t2)
            next_t = self.sample_points.peekitem(next_i)[0]
            prev_t = self.sample_points.peekitem(next_i-1)[0]
            if ( next_t - t2 < 2*jnp.finfo(type(t2)).eps ):
                t2 = next_t
            elif ( t2 - prev_t < 2*jnp.finfo(type(t2)).eps ):
                t2 = prev_t
            else:
                raise NotImplementedError('Conditional subsampling not implemented.')

        # Recall: I(1->3)  = I(1->2) + I(2->3) + dWa(2->3) dWb(1->2)

        I = 0
        w1 = self.sample_points[t1]['w']
        it_lower = self.sample_points.irange(t1, t2)
        it_upper = self.sample_points.irange(t1, t2)
        next(it_upper)
        for t_upper in it_upper:
            t_lower = next(it_lower)
            I += self.sample_points[t_upper]['IToPrevPoint']            
            dw = self.sample_points[t_upper]['w'] - self.sample_points[t_lower]['w']
            dw_from_start = self.sample_points[t_lower]['w'] - w1
            I += dw*dw_from_start

        dw = self.sample_points[t2]['w'] - self.sample_points[t1]['w']
        np.fill_diagonal(I, 0.5*(dw**2 - (t2-t1))) # Diagonal entries work differently

        return I

    def ensure_sample_point(self,t):
        '''
        Checks if sample point exists, adds new sample if necessary

        Parameters
        ----------
        t : float
            Timestamp to potentially add to sample history

        '''

        if not (t>=0):
            raise ValueError('Illegal timestamps for sampling. (Negative?)')
        if t in self.sample_points:
            return

        t_max = self.sample_points.keys()[-1]

        # Compare Kloden-Platen (10.3.7), dimension = d, noiseterms = m
        Delta = t - t_max

        # dW = np.sqrt(Delta)*np.array([next(self.normal_generator) for x in range(0, self.noiseterms)]) # slow :<
        dW = np.sqrt(Delta)* self.normal_generator.get_sample(self.noiseterms,n=self.noiseterms)
        xi = (1.0 / np.sqrt(Delta) * dW).reshape(1, -1)
        # mu = np.random.normal(size=self.noiseterms).reshape(1, -1)
        mu = self.normal_generator.get_sample(size=(1,self.noiseterms),n=self.noiseterms)
        # eta = np.random.normal(size=(self.p, self.noiseterms))
        eta = self.normal_generator.get_sample(size=(self.p,self.noiseterms),n=self.p*self.noiseterms)
        # zeta = np.random.normal(size=(self.p, self.noiseterms))
        zeta = self.normal_generator.get_sample(size=(self.p,self.noiseterms),n=self.p*self.noiseterms)
        rec = 1/np.arange(1, self.p+1) # 1/r vector
        rho = 1/12 - (rec**2).sum()/(2*np.pi**2)

        a = math.sqrt(2)*xi+eta

        Imat = (
            Delta*(xi*xi.T/2 + np.sqrt(rho)*(mu*xi.T - xi*mu.T))
            + Delta/(2*np.pi)*  ((np.expand_dims(a, 1)*np.expand_dims(zeta, 2) - np.expand_dims(a, 2)*np.expand_dims(zeta, 1)).T*rec).sum(axis=-1)
        )
        
        np.fill_diagonal(Imat, 0.5*(dW**2 - Delta)) # Diagonal entries work differently

        self.sample_points[t] = {'w': self.sample_points[t_max]['w'] + dW, 'IToPrevPoint' : Imat}



def f(p, Delta, dW, xi, mu, eta, zeta):
    # not yet verified vectorization for VectorWienerWithI.ensure_sample_point
    rec = 1/jnp.arange(1, p+1) # 1/r vector
    rho = 1/12 - (rec**2).sum()/(2*jnp.pi**2)

    a = jnp.sqrt(2)*xi+eta
    muxi = mu*xi.T
    azeta1 = jnp.expand_dims(a, 1)*jnp.expand_dims(zeta, 2)
    azeta2 = jnp.transpose(azeta1, (0, 2, 1))
    Imat = (
        Delta*(xi*xi.T/2 + np.sqrt(rho)*(muxi - muxi.T))
        + Delta/(2*np.pi)*  ((azeta1 - azeta2).T*rec).sum(axis=-1)
    )
    
    Imat = fill_diagonal(Imat, 0.5*(dW**2 - Delta)) # Diagonal entries work differently

def fill_diagonal(a, val):
  assert a.ndim >= 2
  i, j = jnp.diag_indices(min(a.shape[-2:]))
  return a.at[..., i, j].set(val)